////////////////////////////////////////////////////////////////////////////////
//
// Except32.h - Delphi like exception handling for C++ 
// ---------------------------------------------------
// Version:   2001-04-14
// Maintain:  Michael Vinther, s973971@student.dtu.dk
//
//
// Last changes:
//   try-finally added
//
#ifndef EXCEPT32_INCLUDED
#define EXCEPT32_INCLUDED
           
#include <windows.h>
#include <strings.h>

// For delphi like try-finally statements
#define finally(FinalizationCode) catch(...) { FinalizationCode throw; } FinalizationCode

// Basic exception with message
class Exception
  {
    public:
      TString Message;

      Exception(TString &message)
      {
        Message=message;
      }
      Exception(char *message)
      {
        Message=message;
      }
  };

class EWin32Error : public Exception
  {
    public:
      unsigned int ErrorCode;

      EWin32Error(unsigned int Code, TString Message) : Exception(Message)
      {
        ErrorCode=Code;
      }
      EWin32Error(unsigned int Code, char *Message) : Exception(Message)
      {
        ErrorCode=Code;
      }
  };


// Get Win32 error message string
TString SysErrorMessage(int ErrorCode)
{
  int Len;
  char Buffer[256];

  Len=FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ARGUMENT_ARRAY,
                    0, ErrorCode, 0, Buffer, sizeof(Buffer), 0)-1;
  while(Len>0 && (Buffer[Len]==0 || Buffer[Len]==32 || Buffer[Len]==13 || Buffer[Len]==10 || Buffer[Len]=='.')) Len--;
  Buffer[Len+1]=0;

  return TString(Buffer);
}

// Get last Win32 error and throw exceptin if not ERROR_SUCCESS
void ThrowLastWin32Error()
{
  DWORD LastError;

  LastError = GetLastError();
  if(LastError!=ERROR_SUCCESS) throw EWin32Error(LastError,SysErrorMessage(LastError));
}

#endif