////////////////////////////////////////////////////////////////////////////////
//
// Strings.h - String handling routines
// ------------------------------------
// Version:   2002-05-14
// Maintain:  Michael Vinther, mv@logicnet.dk
//
//
// Last changes:
//  == operator overloaded, UpperCase function

#ifndef STRINGS_INCLUDED 
#define STRINGS_INCLUDED

#include <string.h>
#include <stdlib.h>

const int MaxShortStringLength = 255;

class TShortString {
  protected:
    char fString[MaxShortStringLength+1];
  public:
    // Constructors
    TShortString() { fString[0]=0; }
    TShortString(TShortString &s) { strcpy(fString,s.fString); }
    TShortString(const char *s);
    TShortString(int Value) { itoa(Value,fString,10); }

    // Return string length
    int Length() { return strlen(fString); }

    char& operator[](int n) { return fString[n]; }

    operator char*() { return fString; }

    TShortString operator+(TShortString &Tail);
    TShortString operator+(char *s);
    void operator+=(char *s);
    bool operator==(char *s);
};

// Overload + operator
TShortString operator+(char *s1, TShortString s2) {
  return TShortString(s1)+s2;
}

#define TString TShortString

// Pascal-like Pos function. Returns -1 if SubStr not found
int Pos(char *SubStr, char *Str) {
  char *Result = strstr(Str,SubStr);
  if(Result) return Result-Str;
  else return -1;
}

char UpCase(char ch)
{
  if(ch>='a' && ch<='z') 
    ch &= ~('A' ^ 'a');
  return ch;
}

char DnCase(char ch)
{
  if(ch>='A' && ch<='Z') 
    ch |= ('A' ^ 'a');
  return ch;
}

TString UpperCase(char *s)
{
  TString Str = s;
  s = Str;
  while(*s) {
    if(*s>='a' && *s<='z') 
      *s &= ~('A' ^ 'a');
    s++;
  }
  return Str;
}      

//==============================================================================
// TShortString
//==============================================================================

TShortString::TShortString(const char *s)
{
  int CopyCount = strlen(s)+1;
  if(CopyCount>MaxShortStringLength) {
    CopyCount = MaxShortStringLength;
    fString[MaxShortStringLength]=0;
  }

  memcpy(fString,s,CopyCount);
}

TShortString TShortString::operator+(TShortString &Tail)
{
  TShortString Result;
  strcpy(Result.fString,fString);

  int Len=Length();
  int CopyCount = Tail.Length()+1;
  if(Len+CopyCount>MaxShortStringLength) {
    CopyCount = MaxShortStringLength-Len;
    Result.fString[MaxShortStringLength]=0;
  }

  memcpy(&Result.fString[Len],Tail.fString,CopyCount);

  return Result;
}

TShortString TShortString::operator+(char *s)
{
  TShortString Result;
  strcpy(Result.fString,fString);

  int Len=Length();
  int CopyCount = strlen(s)+1;
  if(Len+CopyCount>MaxShortStringLength) {
    CopyCount = MaxShortStringLength-Len;
    Result.fString[MaxShortStringLength]=0;
  }

  memcpy(&Result.fString[Len],s,CopyCount);

  return Result;
}

void TShortString::operator+=(char *s)
{
  int Len=Length();
  int CopyCount = strlen(s)+1;
  if(Len+CopyCount>MaxShortStringLength) {
    CopyCount = MaxShortStringLength-Len;
    fString[MaxShortStringLength]=0;
  }

  memcpy(&fString[Len],s,CopyCount);
}

bool TShortString::operator==(char *s)
{
  return strcmp(fString,s)==0;
}

#endif