////////////////////////////////////////////////////////////////////////////////
//
// DirFile.c - Direct disk access
// ------------------------------
// Version:   2003-12-28
// Maintain:  Michael Vinther   |   mv@logicnet.dk
//
// Last changes:
//   External stream mode
//
#include <stdio.h>

#ifndef DirectFile_INCLUDED
#define DirectFile_INCLUDED

#include "dirfile.h"
#include <stdlib.h>
#include <malloc.h>
#include <Winioctl.h>

#define DirectFileBufferSize 2048
#define DirectStreamCacheSize (1024*1024)

PDirectFile DirectFileOpen(char *FileName,__int64 Offset,__int64 Size)
{
  HANDLE Handle = CreateFile(FileName,GENERIC_READ,0,0,OPEN_EXISTING,FILE_FLAG_NO_BUFFERING,0);

  if(Handle==INVALID_HANDLE_VALUE)
    return 0;
  else {
    PDirectFile File = (PDirectFile)malloc(sizeof(TDirectFile));

    if(Size<0) {
      PDWORD PSize = (PDWORD)&Size;
      *PSize = GetFileSize(Handle,PSize+1);
      Size-=Offset;
    }

    File->Handle = Handle;
    File->StreamGet = NULL;
    File->Buffer = (char*)VirtualAlloc(0,DirectFileBufferSize,MEM_COMMIT,PAGE_READWRITE);
    File->Offset = Offset;
    File->Size = Size;
    File->BufferSize = 0;
    File->BufferPos = 0;

    if(Offset) {
      File->Position = -1;
      DirectFileSeek(File,0,SEEK_SET);
    }
    else File->Position = 0;

    return File;
  }
}

PDirectFile PhysicalDriveOpen(char Drive,__int64 Offset,__int64 Size)
{
  char FileName[] = "\\\\.\\PHYSICALDRIVE0";
  FileName[strlen(FileName)-1] = Drive+'0';

  if(Size<0) {
    PDirectFile File = DirectFileOpen(FileName,0,0);

    if(File) {
      __int64 Size;

      DISK_GEOMETRY DiskInfo;
      DWORD Returned;

      // Get disk size from DeviceIoControl
      DeviceIoControl(File->Handle,IOCTL_DISK_GET_DRIVE_GEOMETRY,NULL,0,&DiskInfo,sizeof(DiskInfo),&Returned,0);
      Size = (__int64)(DiskInfo.BytesPerSector)*DiskInfo.Cylinders.QuadPart*DiskInfo.TracksPerCylinder*DiskInfo.SectorsPerTrack;

      /*PDWORD PSize = (PDWORD)&Size; // Get size by seeking to end
      Size = 0;
      *PSize = SetFilePointer(File->Handle,*PSize,(PLONG)(PSize+1),FILE_END);
      if(*PSize=-1) ThrowLastWin32Error();*/

      File->Size = Size-Offset;
      if(Offset) {
        File->Offset = Offset;
        File->Position = -1;
        DirectFileSeek(File,0,SEEK_SET);
      }
      else SetFilePointer(File->Handle,0,0,FILE_BEGIN);
    }
    return File;
  }
  else return DirectFileOpen(FileName,Offset,Size);
}

PDirectFile DirectStreamOpen(TStreamGetCallback StreamGetCallback,__int64 Size,int SequentialStream,DWORD ID)
{
  PDirectFile File = (PDirectFile)malloc(sizeof(TDirectFile));

  if(Size<0) 
    Size = 0x6fffffffffffffff;

  File->Handle = 0;
  File->StreamGet = StreamGetCallback;
  File->Offset = 0;
  File->Size = Size;
  File->Position = 0;
  if(SequentialStream) {
    File->Buffer = (char*)VirtualAlloc(0,DirectStreamCacheSize,MEM_COMMIT,PAGE_READWRITE);
    File->BufferSize = 0;
    File->BufferPos = 0;
    File->BufferOffset = 0;
  }
  else {
    File->Buffer = NULL;
    File->BufferSize = -1;
  }
  File->ID = ID;
  return File;
}

int DirectFileClose(PDirectFile File)
{
  if(File) {
    int Result = 1;
    if(File->Handle) Result = CloseHandle(File->Handle);
    if(File->Buffer) VirtualFree(File->Buffer,0,MEM_RELEASE);
    free(File);
    return !Result;
  }
  else
    return -1;
}

DWORD DirectFileRead(PDirectFile File,char *Buffer,DWORD Count)
{
  DWORD BlockSize, Result = 0;
  if(File->Position+Count>File->Size)
    Count = (DWORD)(File->Size-File->Position);

  if(File->Handle) { // Disk file
    while(Result<Count) {
  
      if(File->BufferPos>=File->BufferSize) {
        ReadFile(File->Handle,File->Buffer,DirectFileBufferSize,&File->BufferSize,0);
        if(!File->BufferSize) break;
        File->BufferPos = 0;
      }
  
      BlockSize = min(Count-Result,File->BufferSize-File->BufferPos);
      memcpy(Buffer,File->Buffer+File->BufferPos,BlockSize);
      Buffer += BlockSize;
      File->BufferPos += BlockSize;
      Result += BlockSize;
    }
  }
  else { // Stream
    if(File->BufferSize==-1) {
      Result = File->StreamGet(File->Position,Buffer,Count,File->ID); // Stream owner handles to seeking
    }
    else {
      //File->BufferOffset+File->BufferSize always points to the currest position of the source stream

      __int64 Get;
      __int64 PosInCache = File->Position-File->BufferOffset;
      int ReadTo;

      if(PosInCache<0) { // Request is before start of cache, reset
        PosInCache = 0;
        File->BufferOffset = File->Position;
        File->BufferSize = 0;
        File->BufferPos = 0;
      }
      else while(PosInCache>File->BufferSize) { // Request is after end of cache, catch up

        ReadTo = File->BufferPos+File->BufferSize;
        if(ReadTo>=DirectStreamCacheSize)
          ReadTo -= DirectStreamCacheSize;        
          
        Get = min(PosInCache-File->BufferSize,DirectStreamCacheSize-ReadTo);
        Get = File->StreamGet(File->BufferOffset+File->BufferSize,File->Buffer+ReadTo,Get,File->ID);

        if(Get==0) return 0;

        File->BufferSize += Get;
        if(File->BufferSize>DirectStreamCacheSize) {
          File->BufferOffset += (File->BufferSize-DirectStreamCacheSize);
          File->BufferPos    += (File->BufferSize-DirectStreamCacheSize);
          if(File->BufferPos>=DirectStreamCacheSize)
            File->BufferPos = 0;
          PosInCache = File->Position-File->BufferOffset;
          File->BufferSize = DirectStreamCacheSize;
        }
      }

      Result = 0;
      if(PosInCache<File->BufferSize) { // Use cached data

        int Offset = File->BufferPos+PosInCache; // Start position
        if(Offset>=DirectStreamCacheSize)
          Offset -= DirectStreamCacheSize;

        Get = min(Count,File->BufferSize-PosInCache); // Number of bytes we want from cache

        if(Get>DirectStreamCacheSize-Offset) { // Split cache read into two
          Result = DirectStreamCacheSize-Offset;
          memcpy(Buffer,File->Buffer+Offset,Result);
          Get -= Result;
          Offset = 0;
        }
        memcpy(Buffer+Result,File->Buffer+Offset,Get);
        Result += Get;
      }

      while(Result<Count) { // Read from source stream and fill cache

        ReadTo = File->BufferPos+File->BufferSize;
        if(ReadTo>=DirectStreamCacheSize)
          ReadTo -= DirectStreamCacheSize;        
        
        Get = min(Count-Result,DirectStreamCacheSize-ReadTo); // Number of bytes we want from stream

        Get = File->StreamGet(File->BufferOffset+File->BufferSize,File->Buffer+ReadTo,Get,File->ID);

        if(Get==0) break;

        memcpy(Buffer+Result,File->Buffer+ReadTo,Get);
        Result += Get; 
        File->BufferSize += Get;
        if(File->BufferSize>DirectStreamCacheSize) {
          File->BufferOffset += (File->BufferSize-DirectStreamCacheSize);
          File->BufferPos    += (File->BufferSize-DirectStreamCacheSize);
          if(File->BufferPos>=DirectStreamCacheSize)
            File->BufferPos = 0;
          File->BufferSize = DirectStreamCacheSize;
        }
      }
    }
  }
  File->Position+=Result;

  return Result;
}

__int64 DirectFileSeek(PDirectFile File,__int64 Offset, int Origin)
{
  __int64 SeekTo, NewPosition;
  PDWORD PSeekTo = (PDWORD)&SeekTo;

  switch(Origin) {
    case SEEK_SET:
      NewPosition = Offset;
      break;
    case SEEK_CUR:
      NewPosition = File->Position+Offset;
      break;
    case SEEK_END:
      NewPosition = File->Size+Offset;
      break;
    default:
      return File->Position;
  };

  if(NewPosition>File->Size) NewPosition = File->Size;

  if(NewPosition==File->Position) return NewPosition; // Check for no movement

  if(File->Handle) { // Disk file
    NewPosition+=File->Offset;
    SeekTo = NewPosition;
    *PSeekTo = *PSeekTo & (0xffffffff-(DirectFileBufferSize-1)); // Round down to buffer size
  
    *PSeekTo = SetFilePointer(File->Handle,*PSeekTo,(PLONG)(PSeekTo+1),FILE_BEGIN);
  
    File->BufferPos = (DWORD)(NewPosition-SeekTo);
  
    if(File->BufferPos)
      ReadFile(File->Handle,File->Buffer,DirectFileBufferSize,&File->BufferSize,0);
    else
      File->BufferSize = 0;
  }
  else { // Stream
  
  }    

  File->Position = NewPosition-File->Offset;
  return File->Position;
}

__int64 DirectFilePosition(PDirectFile File)
{
  return File->Position;
}

__int64 DirectFileSize(PDirectFile File)
{
  return File->Size;
}

#endif
