////////////////////////////////////////////////////////////////////////////////
//
// MPEG2Lib.dll project
// --------------------
// Version:   2004-01-10
// Maintain:  Michael Vinther   |   mv@logicnet.dk
//
// Last changes:
//   VideoPTS added

#include "stdafx.h"
#include "mpeg2lib.h"
#include <strings.h>
#include <fstream>
#include <except32.h>

#define GLOBAL // Get global constants
extern "C"
{
#include "dvd2avi/global.h"
#include "mpeg2dec.h"
}

static void Initialize();

class EStopPlaying { };

int GetIframe=0;  // Get first frame without advancing pointer!

BOOL APIENTRY DllMain(HANDLE hModule,DWORD ul_reason_for_call,LPVOID lpReserved)
{
  switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
      Initialize();
      break;
		case DLL_THREAD_ATTACH:
      break;
		case DLL_THREAD_DETACH:
      break;
		case DLL_PROCESS_DETACH:
      DirectFileClose(VideoFile);
      File_Limit=0;
      Recovery();
			break;
  }
  return TRUE;
}

static void PrepareDecoder()
{
  //strcpy(szOutput,"x:\\x"); // Audio stream file
  //Track_Flag = 0; // Select audio track
                  
	process.rightlba = (int)(process.length[0]/BUFFER_SIZE);
	process.end = process.total - BUFFER_SIZE;
	process.endloc = (process.length[0]/BUFFER_SIZE - 1)*BUFFER_SIZE;
	process.locate = LOCATE_INIT;
  try {
    MPEG2Dec(0);
  }
  catch(EStopPlaying) { }
  process.locate = LOCATE_RIP;
}

DLLExport(int) OpenMPEG2File(char *FileName,int64 Offset,int64 Size)
{
  DirectFileClose(VideoFile);
  VideoFile = DirectFileOpen(FileName,Offset,Size);
  if(!VideoFile) return 0;
  File_Limit = 1;
  Recovery();
  if(!process.total) return 0;

  PrepareDecoder();
  GetIframe=1;
  return 1;
}

DLLExport(int) OpenMPEG2Disk(byte Drive,int64 Offset,int64 Size)
{
  DirectFileClose(VideoFile);
  VideoFile = PhysicalDriveOpen(Drive,Offset,Size);
  if(!VideoFile) return 0;
  File_Limit = 1;
  Recovery();
  if(!process.total) return 0;

  PrepareDecoder();
  GetIframe=1;
  return 1;
}

DLLExport(int) OpenMPEG2Stream(TStreamGetCallback StreamGetCallback,int SequentialStream,__int64 Size,DWORD ID)
{
  DirectFileClose(VideoFile);
  VideoFile = DirectStreamOpen(StreamGetCallback,Size,SequentialStream,ID);
  if(!VideoFile) return 0;
  File_Limit = 1;
  Recovery();
  if(!process.total) return 0;

  PrepareDecoder();
  return 1;
}

DLLExport(void) CloseMPEG2File()
{
  if(File_Limit) {
    File_Limit = 0;
    DirectFileClose(VideoFile); VideoFile = 0;
    Recovery();
  }
}

DLLExport(void) SetMPEG2PixelFormat(int Format)
{
  Store_Flag = Format;
}

DLLExport(void) GetMPEG2FrameInfo(TVideoFrameInfo *FrameInfo)
{
  FrameInfo->Width = Coded_Picture_Width;
  FrameInfo->Height = Coded_Picture_Height;
  FrameInfo->FrameRate = frame_rate;
  FrameInfo->AspectRatio= aspect_ratio_information; //2004-10-24: Added to get the aspectratio
}

DLLExport(void) GetMPEG2FileInfo(TVideoFileInfo *FileInfo)
{
  FileInfo->Frame = Frame_Number+1;
  FileInfo->Position = DirectFilePosition(VideoFile);
  FileInfo->Size = DirectFileSize(VideoFile);
  FileInfo->VideoPTS = VideoPTS;
}                                

DLLExport(pbyte) GetMPEG2Frame()
{
  try {
	  if (GetIframe==1) {
          Write_Frame(backward_reference_frame, d2v_backward, 0);
		  GetIframe=0;
		  try {
			//first P-frame decoding
			while (Get_Hdr() && picture_coding_type==B_TYPE);
			Decode_Picture();
		  } catch(EStopPlaying) { }; // catch end of file exception
          return FrameBuffer;
	  }
	  if(Get_Hdr())
	  {
		  Decode_Picture();
	      return FrameBuffer;
	  }
  }
  catch(EStopPlaying) {
	// No more data -> use last P or I frame!
	if (backward_reference_frame) {
      Write_Frame(backward_reference_frame, d2v_backward, 0);
      return FrameBuffer;
	}
  }
  return NULL;
}

DLLExport(void) SkipMPEG2Frames(int FrameCount)
{
  int Format = Store_Flag;

  if (FrameCount<=0) return;

  if (GetIframe==1) {
      //first P-frame decoding
	  while (Get_Hdr() && picture_coding_type==B_TYPE);
	  Decode_Picture();

	  GetIframe=0;
	  FrameCount--;
  }

  Store_Flag = PixelFormatNone;
  try {
    while(FrameCount-- && Get_Hdr())
          Decode_Picture();
  }
  catch(EStopPlaying) { }
  Store_Flag = Format;
}

DLLExport(void) MPEG2Seek(int64 Position)
{
	process.end = process.total-BUFFER_SIZE;
	process.endfile = File_Limit-1;
	process.endloc = (process.length[0]/BUFFER_SIZE-1)*BUFFER_SIZE;

	process.startloc = process.start = Position;

  process.locate = LOCATE_SCROLL;
  try {
    MPEG2Dec(0);
  }
  catch(EStopPlaying) { }
  process.locate = LOCATE_RIP;
  GetIframe=1;
}

DLLExport(void) SetRGBScaleFlag(int DoScaling)
{
	Scale_Flag = DoScaling;
	if (Scale_Flag)
	{
		RGB_Scale = 0x1000254310002543;
		RGB_Offset = 0x0010001000100010;
		RGB_CBU = 0x0000408D0000408D;
		RGB_CGX = 0xF377E5FCF377E5FC;
		RGB_CRV = 0x0000331300003313;
	}
	else
	{
		RGB_Scale = 0x1000200010002000;
		RGB_Offset = 0x0000000000000000;
		RGB_CBU = 0x000038B4000038B4;
		RGB_CGX = 0xF4FDE926F4FDE926;
		RGB_CRV = 0x00002CDD00002CDD;
	}
}

DLLExport(int64) WriteDataToFile(char *FileName, int64 Size)
{
  HANDLE OutFile = CreateFile(FileName,GENERIC_WRITE,0,0,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,0);
  if(OutFile==INVALID_HANDLE_VALUE) 
    return 0;

  char Buffer[64*1024];
  int64 Written = 0;
  int64 StartPosition = DirectFilePosition(VideoFile);

  try {
    DWORD Got = 1;
    while(Size && Got) {
      Got = DirectFileRead(VideoFile,Buffer,sizeof(Buffer));
      WriteFile(OutFile,Buffer,Got,&Got,0);
      Size-=Got;
      Written+=Got;
    }
  }
  finally (
    CloseHandle(OutFile);
  );
  DirectFileSeek(VideoFile,StartPosition,SEEK_SET);
  return Written;
}

void Initialize()
{
  GetCPUInfo();

  // Set default option flags
	if(cpu.ssemmx)
		iDCT_Flag = IDCT_SSEMMX;
	else if(cpu.mmx)
		iDCT_Flag = IDCT_MMX;
  else
    iDCT_Flag = IDCT_FPU;
  File_Limit = 0;
  Store_Flag = PixelFormatRGB24;
	FO_Flag = FO_NONE;
  //FO_Flag = FO_SWAP;
	Track_Flag = TRACK_NONE;
	Format_Flag = FORMAT_AUTO;
	AC3_Flag = AUDIO_DECODE;
	DRC_Flag = DRC_NORMAL;
	DSDown_Flag = false;
	MPA_Flag = AUDIO_DEMUXONE;
	SRC_Flag = SRC_NONE;
	Norm_Ratio = 100;
	KeyOp_Flag = 0;
  SetRGBScaleFlag(false);

  // Prepare temporary buffers
	for (int i=0; i<8; i++)
		block[i] = (short *)_aligned_malloc(sizeof(short)*64, 64);                                                                
	fTempArray = (void *)_aligned_malloc(sizeof(float)*128, 64);

  // Initialize IDCT
	Initialize_REF_IDCT();
	Initialize_FPU_IDCT();
}

void ThreadKill()
{
  if(Fault_Flag>=CRITICAL_ERROR_LEVEL) {
    Fault_Flag = 0;    
    throw EStopPlaying();
  }
}
